from sys import argv
import sys
import os

from epic2.src.reads_to_bins import files_to_bin_counts
from epic2.src.SICER_stats import compute_score_threshold
from epic2.src.find_islands import find_islands, compute_fdr, write_islands, add_chip_count_to_islands
from epic2.src.genome_info import egl_and_chromsizes

from collections import OrderedDict

def _main(args):

    args["drop_duplicates"] = int(not args["keep_duplicates"])

    import logging

    if args["quiet"]:
        level = logging.CRITICAL
    else:
        level = logging.INFO

    logging.basicConfig(
        level=level,
        format=
        '%(message)s ',
        datefmt='%a, %d %b %Y %H:%M:%S',
        stream=sys.stderr)

    output = args["output"]
    if output:
        outfile_directory = os.path.dirname(output)
        if outfile_directory and not os.path.exists(outfile_directory):
            os.makedirs(outfile_directory)

    effective_genome_length, chromsizes = egl_and_chromsizes(args)
    args["chromsizes_"] = chromsizes
    args["effective_genome_size"] = effective_genome_length

    c_bins_counts, chip_count_before = files_to_bin_counts(args["treatment"], args, "ChIP")
    chip_count = sum(sum(counts) for _, counts in c_bins_counts.values())

    logging.info("\nValid ChIP reads: {}\n".format(chip_count))

    score_threshold, island_enriched_threshold, average_window_readcount = compute_score_threshold(
        chip_count, args["bin_size"], effective_genome_length, args["gaps_allowed"] * args["bin_size"], args["e_value"])

    logging.info("Number of tags in a window: {}\n".format(island_enriched_threshold))

    islands = find_islands(c_bins_counts, args["gaps_allowed"], args["bin_size"], score_threshold, island_enriched_threshold, average_window_readcount)
    new_islands = add_chip_count_to_islands(islands, c_bins_counts)

    logging.info("Number of islands found: {}\n".format(sum(len(i) for i in islands.values())))

    if args["control"]:

        b_bins_counts, background_count_before = files_to_bin_counts(args["control"], args, "Input")
        background_count = sum(sum(counts) for _, counts in b_bins_counts.values())
        logging.info("\nValid background reads: {}\n".format(background_count))

        if args["original_algorithm"]:
            cc, bc = chip_count_before, background_count_before
        else:
            cc, bc = chip_count, background_count

        compute_fdr(new_islands, b_bins_counts, chip_count, background_count, effective_genome_length, args["false_discovery_rate_cutoff"], args["output"])
    else:
        write_islands(new_islands, average_window_readcount, args["false_discovery_rate_cutoff"], args["e_value"], args["output"])

        if args["original_algorithm"]:
            cc = chip_count_before
        else:
            cc = chip_count

    return cc



import sys
import logging
from math import pi, log, exp, ceil, fabs

class Background_island_probscore_statistics:
	#  External genomeLength and gapSize are in units of bps
	#  Internal genomeLength and gapSize are in units of windows
	#  only look at enrichment!
	def __init__(self, total_tags, windowSize, gapSize, window_pvalue, genomeLength, bin_size):
		self.tag_density = total_tags * 1.0 / genomeLength;
		self.window_size = windowSize; # In bps.
		assert(gapSize%windowSize == 0); # gap is in bps
		self.gap_size = gapSize/windowSize; # maximum number of windows allowed in a gap
		self.genome_length = int(ceil(float(genomeLength)/windowSize));
		self.average = self.tag_density * windowSize;
		self.bin_size = bin_size;

		# Precalculate the poisson, cumulative poisson values up to max (500, 2*self.average) .
		self.max_index = max (500, int(2*self.average)) ;
		#print self.average, self.max_index;
		#self.fact=[];
		self.poisson_value=[];
		self.window_score=[];
		self.window_scaled_score=[];
		for index in xrange(self.max_index):
			# self.fact.append(self.factorial(index));
			prob = self.poisson(index, self.average);
			self.poisson_value.append(prob);
			if ( index < self.average): # only want to look at enrichment
				self.window_score.append(0);
				self.window_scaled_score.append(0);
			else:
				if prob > 0:
					self.window_score.append(-log(prob));
					#scaled_score =int(-log(prob)/self.bin_size);
					scaled_score =int(round(-log(prob)/self.bin_size))
					self.window_scaled_score.append(scaled_score);
				else: #prob is too small and deemed 0 by the system
					self.window_score.append(1000);
					scaled_score =int(round(1000/self.bin_size))
					self.window_scaled_score.append(scaled_score);
			#print index, self.poisson_value[index], self.window_score[index];
		self.max_index = len(self.poisson_value);
		#print "max_index ", self.max_index;
		# gap_contribution needs min_tags_in_window
		# So the position of this line is critical.
		self.min_tags_in_window = 0;
		sf = 1 ;

		# print("self.average=", self.average)
		# print("self.poisson_value[0]=", self.poisson_value[0])
		# print("window_pvalue", window_pvalue)
		#print "self.poisson_value[0]=", self.poisson_value[0];
		while (sf > window_pvalue ):
			#print self.min_tags_in_window, sf;
			# print("poisson", self.min_tags_in_window, self.poisson_value[self.min_tags_in_window])
			sf -= self.poisson_value[self.min_tags_in_window]
			self.min_tags_in_window += 1;
		#An alternative approach that uses the scipy package,
		#poisson.sf (n, lambda) = \sum_{i= n+1}^{\infty} p(i, lambda)
		#self.min_tags_in_window = int(self.average);
		#while (scipy.stats.poisson.sf(self.min_tags_in_window-1) > window_pvalue):
		#	self.min_tags_in_window += 1;

		#print "Window read count threshold: ", self.min_tags_in_window;

		self.gap_contribution = self.gap_factor();
		self.boundary_contribution = self.boundary();

		# print("self.gap_contribution", self.gap_contribution)
		# print("self.boundary_contribution", self.boundary_contribution)

		self.cumulative=[];
		# new method, first fill the lowest score.
		prob = self.boundary_contribution * self.poisson_value[self.min_tags_in_window];
		score = -log(self.poisson_value[self.min_tags_in_window]);
		#scaled_score = int(score/self.bin_size);
		scaled_score = int(round(score/self.bin_size));
		self.island_expectation =[0] * (scaled_score+1);

		# print("island_expectation length", (scaled_score+1))
		self.island_expectation[scaled_score] = prob*self.genome_length;
		# print("self.island_expectation[scaled_score]", self.island_expectation[scaled_score])
		# print("scaled_score", scaled_score)

		self.island_expectation[0] = self.boundary_contribution*self.genome_length/self.gap_contribution;


		self.root = self.find_asymptotics_exponent()
		#print "Exponent for Asymptotics: ", self.root;

	def factorial(self, m):
		value = 1.0;
		if m != 0:
			while m != 1:
				value = value*m;
				m = m - 1;
		return value;


	# Return the log of a factorial, using Srinivasa Ramanujan's approximation
	def factln(self, m):
		if m<20:
			value = 1.0;
			if m != 0:
				while m != 1:
					value = value*m;
					m = m - 1;
			return log(value);
		else:
			return m*log(m) -m + log(m*(1+4*m*(1+2*m)))/6.0 + log(pi)/2;


	def poisson(self, i, average):
		if i<20:
			return exp(-average) * average**i / self.factorial(i);
		else:
			exponent = -average + i*log(average) - self.factln(i);
			return exp(exponent);

	"""
		gap is in the unit of windows. In each window in the gap, the
		window could have 0, 1, min_tags_in_windows-1 tags.

		say gap = 1, min_tags_in_window= 2, gap_factor = 1 +
		poission(0,a) + poisson(1, a), where 1 represents no gap,
		poisson(0,a) represents a window with 0 tag,
		poisson(1,a) represents a window with 1 tag,

		The gap contribution from each window is not independent
	"""
	def single_gap_factor(self):
		my_gap_factor=0;
		for i in xrange(self.min_tags_in_window):
			my_gap_factor +=self.poisson_value[i];
		return my_gap_factor;

	# gap contribution is bigger than 1
	def gap_factor(self):
		if self.gap_size == 0:
			return 1;
		else:
			i = 1;
			gap_contribution = 1; # contribution from no gap
			my_gap_factor = self.single_gap_factor();
			# print("self.gap_size " * 100, self.gap_size)
			for i in range(1, int(self.gap_size+1)): gap_contribution += pow(my_gap_factor, i);
			return gap_contribution;

	def boundary(self):
		"""
		The condition for boundary is a continuous region of
		unqualified windows longer than gap
		"""
		temp = self.single_gap_factor();
		temp = pow(temp, self.gap_size+1);
		return temp*temp; # start & end

	#forward method that memorize the calculated results.
	def background_island_expectation (self, scaled_score):
		current_max_scaled_score = len(self.island_expectation)-1;
		if scaled_score > current_max_scaled_score:
			#index is the scaled_score
			for index in range(current_max_scaled_score + 1, scaled_score+1):
				temp=0.0;
				#i is the number of tags in the added window
				i = self.min_tags_in_window;
				while ( int(round(index - self.window_score[i]/self.bin_size))>=0):
				#while ( (index - self.window_scaled_score[i])>=0):
					temp += self.poisson_value[i]* self.island_expectation[int(round(index - self.window_score[i]/self.bin_size))];
					#temp += self.poisson_value[i]* self.island_expectation[index - self.window_scaled_score[i]];
					i += 1;
				temp *= self.gap_contribution;
				self.island_expectation.append(temp);
				#print index, temp, self.island_expectation[index];
		return self.island_expectation[scaled_score];


	def generate_cumulative_dist(self,  outfile=""):
		"""
		Generate cumulative distribution: a list of tuples (bins, hist).
		"""
		self.cumulative=[0]*len(self.island_expectation);
		partial_sum = 0.0
		for index in range(1, len(self.island_expectation)+1):
			complimentary = len(self.island_expectation) - index;
			partial_sum += self.island_expectation[complimentary]; # The end is outside of the index
			self.cumulative[complimentary]=partial_sum;

		if outfile != "":
			fixpoint = int(len(self.island_expectation)/2);
			outf = open(outfile, "w");
			outline ="# Score" + "\t" + "Expect # islands" +"\t" + "Cumulative # Islands" + "\t" + "Asymptotics"+ "\n";
			outf.write(outline);
			for index in xrange(len(self.island_expectation)):
				outline = str(index * self.bin_size) + "\t" + str(self.island_expectation[index])+ "\t" +str(self.cumulative[index])  + "\n";
				#outline = str(index * self.bin_size) + "\t" + str(self.island_expectation[index])+ "\t" +str(self.cumulative[index]) + "\t" + str(self.cumulative[fixpoint] * exp(-self.root*(self.cumulative[index]-self.cumulative[fixpoint]))) + "\n";
				outf.write(outline);
			outf.close();


	def find_island_threshold(self, e_value_threshold):
		"""
		average is the average number of tags in a window:
		opt.tag_density * opt.window_size

		This one allows single-window islands.
		Returns the island threshold
		"""
		threshold = .0000001*e_value_threshold;
		current_scaled_score = len (self.island_expectation) - 1;
		current_expectation = self.island_expectation[-1];
		assert (current_expectation == self.island_expectation[current_scaled_score]);
		interval = int(1/self.bin_size);
		if len(self.island_expectation) > interval:
			partial_cumu = sum(self.island_expectation[-interval: -1])
		else:
			partial_cumu = sum(self.island_expectation)
		while ( partial_cumu > threshold or  partial_cumu <1e-100):
			current_scaled_score += interval;
			current_expectation=self.background_island_expectation(current_scaled_score);
			if len(self.island_expectation) > interval:
				partial_cumu = sum(self.island_expectation[-interval: -1])
			else:
				partial_cumu = sum(self.island_expectation)
			#for index in  xrange(len(self.island_expectation)):
					#print  index*self.bin_size, self.island_expectation[index];

		self.generate_cumulative_dist();
		for index in xrange(len(self.cumulative)):
			if self.cumulative[index]<=e_value_threshold:
				score_threshold = index*self.bin_size;
				break;
		return score_threshold;

	def func(self, x):
		sum_ = 0.0;
		for index in range(self.min_tags_in_window, self.max_index):
			sum_ += self.gap_contribution * pow(self.poisson_value[index], 1-x);
		return sum_-1;

	def bracket_root(self, f, interval, max_iterations=50):
		"""\
		Given a univariate function f and a tuple interval=(x1,x2),
		return a new tuple (bracket, fnvals) where bracket=(x1,x2)
		brackets a root of f and fnvals=(f(x1),f(x2)).
		"""
		GOLDEN = (1+5**.5)/2
		(x1, x2) = interval
		if x1==x2:
			raise Exception("initial interval has zero width")
		elif x2<x1:
			x1, x2 = x2, x1
		f1, f2 = f(x1), f(x2)
		for j in xrange(max_iterations):
			while f1*f2 >= 0:  # not currently bracketed
				if abs(f1)<abs(f2):
					x1 = x1 + GOLDEN*(x1-x2)
				else:
					x2 = x2 + GOLDEN*(x2-x1)
				f1, f2 = f(x1), f(x2)
			return (x1, x2), (f1, f2)
		raise Exception("too many iterations")

	# based on Numerical Recipes, p. 354
	def bisect_root(self, func, interval, xacc):
		JMAX=50;
		(x1, x2) = interval;
		f=func(x1);
		fmid=func(x2);
		if (f*fmid >= 0.0): print("Root must be bracketed for bisection");
		if (f < 0.0):
			dx= x2 - x1;
			rtb= x1;
		else:
			dx= x1 - x2;
			rtb= x2;
		for  j in xrange(JMAX):
			dx *= 0.5;
			xmid= rtb + dx;
			fmid=func(xmid);
			if (fmid <= 0.0): rtb=xmid;
			if (fabs(dx) < xacc or fmid == 0.0): return rtb;
		print("Too many bisections");
		return 0.0;

	def find_asymptotics_exponent(self, xacc=.00001):
		num = 100;
		#for index in xrange(num):
		#	x = index/float(num);
		#	print x, self.func(x);
		input_bracket = (0.1 , 1);
		(xresult, yresult) = self.bracket_root(self.func, input_bracket);
		root = self.bisect_root(self.func, xresult, xacc);
		#print "# The exponent is: ", root;
		return root;

# Species:  hg38
# Window_size:  200
# Gap size:  600
# E value is: 1000.0
# Total read count: 20227554.0
# Genome Length:  3088286401
# Effective genome Length:  2625043440
# Window average: 1.54112146807
# opt.evalue: 1000.0
# Total read count: 20227554.0
# window_size: 200
# opt.gap: 600
# Window pvalue: 0.2
# genome_length: 2625043440
# ('self.average=', 1.541121468069877)
# ('self.poisson_value[0]=', 0.2141408146291121)
# ('window_pvalue', 0.2)
# ('poisson', 0, 0.2141408146291121)
# ('poisson', 1, 0.33001700661489664)
# ('poisson', 2, 0.25429814686118796)
# ('poisson', 3, 0.13063477780605437)
# ('self.gap_contribution', 3.5943004243263443)
# ('self.boundary_contribution', 0.5552199656493835)
# ('island_expectation length', 2990)
# ('self.island_expectation[scaled_score]', 366781.3884759084)
# ('scaled_score', 2989)
# Minimum num of tags in a qualified window:  4
# Generate the enriched probscore summary graph and filter the summary graph to get rid of ineligible windows
# Determine the score threshold from random background
# opt.evalue 1000.0
# The score threshold is:  19.112
# Make and write islands
# Total number of islands:  74071


# Find candidate islands exhibiting clustering ...
# /mnt/work/endrebak/software/anaconda/envs/py27/bin/python /home/endrebak/code/epic_paper/SICER/src/find_islands_in_pr.py -s hg38 -b /mnt/scratch/projects/epic_bencmarks/data/sicer_results/0h_exp1/no_bigwig/0h_exp1_chip-W200.graph -w 200 -g 600 -t 0.85 -e 1000 -f /mnt/scra
# tch/projects/epic_bencmarks/data/sicer_results/0h_exp1/no_bigwig/0h_exp1_chip-W200-G600.scoreisland
# Species:  hg38
# Window_size:  200
# Gap size:  600
# E value is: 1000.0
# Total read count: 2792406.0
# Genome Length:  3088286401
# Effective genome Length:  2625043440
# Window average: 0.212751222128
# opt.evalue: 1000.0
# Total read count: 2792406.0
# window_size: 200
# opt.gap: 600
# Window pvalue: 0.2
# genome_length: 2625043440
# ('self.average=', 0.21275122212834696)
# ('self.poisson_value[0]=', 0.8083572135909076)
# ('window_pvalue', 0.2)
# ('poisson', 0, 0.8083572135909076)
# ('self.gap_contribution', 2.990012655388544)
# ('self.boundary_contribution', 0.18231673960386235)
# ('island_expectation length', 1761)
# ('self.island_expectation[scaled_score]', 411536.58828177786)
# ('scaled_score', 1760)
# Minimum num of tags in a qualified window:  1
# Generate the enriched probscore summary graph and filter the summary graph to get rid of ineligible windows
# ('read_count', 1.0, 'start', 1949600)
# ('read_count', 1.0, 'start', 1949800)
# ('read_count', 1.0, 'start', 1950000)
# ('read_count', 3.0, 'start', 1950200)
# ('read_count', 1.0, 'start', 1950600)
# ('read_count', 1.0, 'start', 1950800)
# ('read_count', 1.0, 'start', 1951000)
# ('read_count', 1.0, 'start', 1951600)
# ('read_count', 1.0, 'start', 1951800)
# ('read_count', 1.0, 'start', 1952600)
# ('read_count', 1.0, 'start', 1953400)
# ('read_count', 1.0, 'start', 1953600)
# ('read_count', 1.0, 'start', 1954000)
# Determine the score threshold from random background
# opt.evalue 1000.0
# The score threshold is:  27.363
# Make and write islands
# Total number of islands:  16431


def compute_score_threshold(chip_counts, window_size, effective_genome_length, gap_size, e_value):

	tag_density = chip_counts/effective_genome_length
	background = Background_island_probscore_statistics(chip_counts, window_size, gap_size, 0.2, effective_genome_length, 0.001)
	score_threshold = background.find_island_threshold(e_value)

	logging.info("Score threshold: {}\n".format(score_threshold))

	min_tags_in_window = background.min_tags_in_window

	average_window_readcount = chip_counts * (window_size / float(effective_genome_length))

	return score_threshold, min_tags_in_window, average_window_readcount




from libcpp.pair cimport pair
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.map cimport map
from libc.stdint cimport uint32_t
# cdef from namespace "std":

ctypedef pair[string, char] key
ctypedef vector[uint32_t] intvec

ctypedef map[key, intvec] genome_map


cdef extern from "read_files.cpp":
    genome_map read_bed(const char *, uint32_t)
    genome_map read_bed_gz(const char *, uint32_t)
    genome_map read_bedpe(const char *, uint32_t)
    genome_map read_bedpe_gz(const char *, uint32_t)
    genome_map read_bam(const char *, uint32_t)




# cdef extern from "../../lib/htslib/htslib/hts.h":
#     hts_open(const char *fn, const char *mode)



# cdef extern from "lib/htslib/htslib/sam.h":
#     ctypedef struct bam_hdr_t:
#         pass

#     ctypedef struct samFile:
#         pass

    pass
    # bam_hdr_t *sam_hdr_read(samFile *fp)
#     bam_init1()

# cdef extern from "SICER2/src/"



# typedef struct {
#     int32_t n_targets, ignore_sam_err;
#     uint32_t l_text;
#     uint32_t *target_len;
#     int8_t *cigar_tab;
#     char **target_name;
#     char *text;
#     void *sdict;
# } bam_hdr_t;



# require pysam is pre-installed
try:
    import pysam
except ImportError:
    raise Exception('\n\n' + 'pysam not found; please install pysam first (pip install pysam)' * 5 + '\n\n')

import os
import sys

from distutils.core import setup

from setuptools import find_packages, Extension, Command
from Cython.Build import cythonize

__version__ = open("epic2/version.py").readline().split(" = ")[1].replace('"', '').strip()
macros = []

install_requires = ["scipy", "numpy", "natsort", "cython", "pysam"]


if os.getenv("TRAVIS"):
    install_requires.append("coveralls")


if sys.version_info[0] == 2:
    install_requires.append("functools32")

compile_options = ["-Ofast", "-Wall", "-std=c++11"] #, "-frename-registers", "-funroll-loops"] # , "-lgzstream", "-lz"


from subprocess import check_output

conda_path = check_output("which conda", shell=True).decode().strip()

conda_include = []
conda_lib = []
if conda_path:
    conda_base = conda_path.replace("bin/conda", "")
    conda_include.append(conda_base + "include/")
    conda_lib.append(conda_base + "lib/")



extensions = [Extension("epic2.src.reads_to_bins",
                        ["epic2/src/reads_to_bins.pyx", "epic2/src/gzstream.cpp"], language="c++",
                        include_dirs=conda_include,
                        library_dirs=conda_lib,
                        extra_compile_args=compile_options,
                        libraries=["z"]),
              Extension("epic2.src.SICER_stats",
                        ["epic2/src/SICER_stats.pyx"], language="c++",
                        extra_compile_args=compile_options),
              Extension("epic2.src.statistics",
                        ["epic2/src/statistics.pyx"], language="c++",
                        extra_compile_args=compile_options),
              Extension("epic2.src.find_islands",
                        ["epic2/src/find_islands.pyx"], language="c++",
                        extra_compile_args=compile_options),
              Extension("epic2.src.read_bam",
                        ["epic2/src/read_bam.pyx"], language="c++",
                        extra_compile_args=compile_options,
                        include_dirs=pysam.get_include()  + conda_include,
                        library_dirs=conda_lib,
                        define_macros=pysam.get_defines(),
                        libraries=["z"]),
              Extension("epic2.src.genome_info",
                        ["epic2/src/genome_info.pyx"], language="c++",
                        extra_compile_args=compile_options)]

setup(
    name="epic2",
    packages=find_packages(),
    ext_modules = cythonize(extensions, annotate=True, language_level='3'),
    scripts=["bin/epic2", "bin/epic2-df"],
    package_data={'epic2': ['effective_sizes/*.txt', 'chromsizes/*chromsizes', 'examples/*.bed.gz'],
                  '': ['*.pyx', '*.pxd', '*.h', '*.c', '*.hpp', "*.bed.gz"]},
    version=__version__,
    description="Ultraperformant ChIP-Seq broad peak/domain finder.",
    author="Endre Bakken Stovner",
    author_email="endrebak85@gmail.com",
    url="http://github.com/endrebak/epic2",
    keywords=["ChIP-Seq"],
    license=["MIT"],
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment", "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Scientific/Engineering"
    ],
    include_dirs=["."],
    long_description=
    ("Chip-Seq broad peak/domain finder based on SICER. See the url for more info."
     ))




from cython.operator import dereference, postincrement
from libc.stdint cimport uint32_t, uint16_t, int32_t, int64_t, uint64_t

from cython.operator import dereference
from libcpp.algorithm cimport sort as stdsort
from libcpp.map cimport map as cppmap
from libcpp.algorithm cimport unique
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp.pair cimport pair
from libcpp.map cimport map

cdef extern from "<utility>" namespace "std" nogil:
    pair[T,U] make_pair[T,U](T&,U&)

ctypedef struct interval:
    uint32_t start
    uint32_t end

ctypedef vector[uint32_t] intvec
ctypedef vector[interval] interval_vector
ctypedef pair [string, char] key
ctypedef pair [int, char] intkey
ctypedef map[intkey, intvec] genome_map_int
ctypedef map[key, intvec] genome_map
ctypedef map[intkey, interval_vector] genome_intervals_int
ctypedef map[key, intvec] genome_intervals


# include "<htslib>"


# cpdef read_bam(filename):

#     cdef:


cdef uint32_t compare_start_end(interval lhs, interval rhs):
  if (lhs.start < rhs.start):
    return <uint32_t> 1
  elif (rhs.start < lhs.start):
      return <uint32_t> 0
  elif (lhs.end < rhs.end):
      return <uint32_t> 1
  else:
    return <uint32_t> 0


# uint32_t compare_by_start_end(const interval lhs, const interval rhs){
#   if (lhs.start < rhs.start){
#     return 1;
#   } else if (rhs.start < lhs.start){
#     return 0;
#   } else if (lhs.end < rhs.end){
#     return 1;
#   } else {
#     return 0;
#   };

# }

cdef uint32_t start_end_equal(interval lhs, interval rhs):
  if ((lhs.start == rhs.start) and (lhs.end == rhs.end)):
      return <uint32_t> 1
  else:
      return <uint32_t> 0




# uint32_t start_end_equal(const interval lhs, const interval rhs){
#   if ((lhs.start == rhs.start) && (lhs.end == rhs.end)){
#       return 1;
#   } else {
#     return 0;
#   }
# }


import pysam


from pysam.libcalignmentfile cimport AlignmentFile, AlignedSegment


cpdef read_bam(filename, uint32_t drop_duplicates, uint32_t mapq, uint64_t required_flag, uint64_t filter_flag):

    cdef:
        uint32_t flag
        int32_t start
        int32_t end
        int32_t length
        uint32_t is_strand
        char forward = b"+"
        char reverse = b"-"
        char strand
        string chromosome
        int chromosome_id
        intkey chrom_strand
        key chrom_strand_fixed
        uint32_t five_end
        genome_intervals_int genome
        genome_intervals genome_fixed
        interval _interval
        interval_vector intervals
        intvec five_ends
        # AlignmentFile samfile
        # AlignedSegment a
        uint32_t i = 0

    samfile = pysam.AlignmentFile(filename, "rb")


    for a in samfile:
        flag = a.flag

        # https://broadinstitute.github.io/picard/explain-flags.html

        if a.mapping_quality < mapq:
            continue
        if (flag & required_flag) != required_flag:
            continue
        if (flag & filter_flag) != 0:
            continue

        is_reverse = flag & 0x10

        start = a.reference_start

        end = start + a.alen

        if start < 0 or end < 0:
            continue

        _interval = [<uint32_t> (start + 1), <uint32_t> (end + 1)]

        chromosome_id = a.reference_id
        if is_reverse:
            chrom_strand = make_pair(<int>chromosome_id, <char>reverse)
            genome[chrom_strand].push_back(_interval)
        else:
            chrom_strand = make_pair(<int>chromosome_id, <char>forward)
            genome[chrom_strand].push_back(_interval)


    it = genome.begin();

    while (it != genome.end()):
        chromosome_id = dereference(it).first.first
        chromosome = samfile.get_reference_name(chromosome_id).encode("utf-8")

        strand = dereference(it).first.second
        intervals = dereference(it).second
        five_ends = intvec()

        if drop_duplicates:

            stdsort(intervals.begin(), intervals.end(), compare_start_end)
            intervals.erase(unique(intervals.begin(), intervals.end(), start_end_equal), intervals.end())

        if chr(strand) == "+":
            chrom_strand_fixed = make_pair(<string>chromosome, <char>forward)
            for i in range(intervals.size()):
                five_ends.push_back(intervals[i].start)
        else:
            chrom_strand_fixed = make_pair(<string>chromosome, <char>reverse)
            for i in range(intervals.size()):
                five_ends.push_back(intervals[i].end)

        genome_fixed[chrom_strand_fixed] = five_ends

        postincrement(it)

    return genome_fixed



import logging
import gzip
from os.path import basename
from natsort import natsorted
from collections import OrderedDict

from libc.stdint cimport uint32_t, uint16_t
import numpy as np
import sys
from re import search, IGNORECASE
import pkg_resources



def sniff(f):

    try:
        import pysam
        pysam.AlignmentFile(f, "rb")
        return "bam"
    except:
        pass

    try:
        import gzip
        l = gzip.open(f).readline().decode().split()
        if len(l) == 6:
            return "bed.gz"
        elif len(l) == 10:
            return "bedpe.gz"
        else:
            raise ValueError("Wrong number columns in {}: {}.\nLine looks like: {}".format(f, len(l), l))
    except:
        pass

    try:
        l = open(f).readline().split()
        if len(l) == 6:
            return "bed"
        elif len(l) == 10:
            return "bedpe"
        else:
            raise ValueError("Wrong number columns in {}: {}.\nLine looks like: {}".format(f, len(l), l))

    except:
        raise Exception("Could not find format of file {}.".format(f))

def get_closest_readlength(estimated_readlength):
    # type: (int) -> int
    """Find the predefined readlength closest to the estimated readlength.

    In the case of a tie, choose the shortest readlength."""

    readlengths = [36, 50, 75, 100]
    differences = [abs(r - estimated_readlength) for r in readlengths]
    min_difference = min(differences)
    index_of_min_difference = [i
                               for i, d in enumerate(differences)
                               if d == min_difference][0]

    return readlengths[index_of_min_difference]


def find_readlength(args):
    # type: (Namespace) -> int
    """Estimate length of reads based on 1000 first."""

    # from subprocess import check_output
    # import pandas as pd
    # from io import BytesIO
    _file = args["treatment"][0]
    file_format = sniff(_file)

    arr = np.zeros(100, dtype=np.uint32)

    cdef:
        int i = 0
        uint32_t start
        uint32_t end

    if file_format == "bed":
        for line in open(_file):
            _start, _end = line.split()[1:3]
            arr[i] = int(_end) - int(_start)
            i += 1
            if i == 100:
                break


    elif file_format == "bed.gz":
        for line in gzip.open(_file):
            line = line.decode()
            _start, _end = line.split()[1:3]
            arr[i] = int(_end) - int(_start)
            i += 1
            if i == 100:
                break


    elif file_format == "bedpe":
        for line in open(_file):
            ls = line.split()
            _start = ls[1]
            _end = ls[5]
            arr[i] = int(_end) - int(_start)
            i += 1
            if i == 100:
                break

    elif file_format == "bedpe.gz":
        for line in gzip.open(_file):
            line = line.decode()
            ls = line.split()
            _start = ls[1]
            _end = ls[5]
            arr[i] = int(_end) - int(_start)
            i += 1
            if i == 100:
                break


    elif file_format == "bam":

        import pysam

        if _file.endswith(".bam"):
            samfile = pysam.AlignmentFile(_file, "rb")
        elif _file.endswith(".sam"):
            samfile = pysam.AlignmentFile(_file, "r")

        for a in samfile:
            if a.alen == None: continue

            arr[i] = a.alen
            i += 1
            if i == 100:
                break

    else:
        raise IOError("Cannot recognize file extension of: " + _file + ". Must be bed, bedpe, bam, sam, bed.gz or bedpe.gz")

        # print(arr)
    median = np.median(arr)
    logging.info("Found a median readlength of {}\n".format(median))


    return get_closest_readlength(median)



def get_genome_size_file(genome):

    genome_names = pkg_resources.resource_listdir("epic2", "chromsizes")
    name_dict = {n.lower().replace(".chromsizes", ""): n for n in genome_names}

    # # No try/except here, because get_egs would already have failed if genome
    # # did not exist
    genome_exact = name_dict[genome.lower()]

    return pkg_resources.resource_filename(
        "epic2", "chromsizes/{}".format(genome_exact))


def create_genome_size_dict(genome):
    # type: (str) -> Dict[str,int]
    """Creates genome size dict from string containing data."""

    size_file = get_genome_size_file(genome)
    size_lines = open(size_file).readlines()

    size_dict = {}
    for line in size_lines:
        genome, length = line.split()
        size_dict[genome] = int(length)

    return size_dict


def create_genome_size_dict_custom_genome(chromsizes):

    chromosome_lengths = [l.split() for l in open(chromsizes).readlines()]

    od = OrderedDict()          # type: OrderedDict[str, int]

    for c, l in natsorted(chromosome_lengths):
        od[c] = int(l)

    return od


def get_effective_genome_fraction(genome, read_length):

    genome_names = pkg_resources.resource_listdir("epic2",
                                                  "effective_sizes")

    name_dict = {n.lower(): n for n in genome_names}

    # # No try/except here, because get_egs would already have failed if genome
    # # did not exist
    genome_exact = name_dict[genome.lower() + "_" + str(read_length) + ".txt"]


    try:
        egf = pkg_resources.resource_string( # type: ignore
            "epic2", "effective_sizes/{}".format(
                genome_exact)).split()[-1].decode()
    except KeyError:

        genome_list = "\n".join([basename(g) for g in genome_names])

        raise Exception("Genome " + genome +
            " not found.\n These are the available genomes: " + genome_list +
            "\nIf yours is not there, please request it at github.com/endrebak/epic2 .")

    return float(egf)



def get_chroms_from_bam(bamfile, remove_pattern):
    """Read bam file and return dict of {chrom_name: length}
    """
    import pysam
    inbam = pysam.AlignmentFile(bamfile)
    lengths = inbam.lengths
    references = inbam.references
    chroms = {}

    removed = []

    if not remove_pattern:
        for x, l in zip(references, lengths):
            chroms[x] = l
    else:
        import re
        pattern = re.compile(remove_pattern)
        for x, l in zip(references, lengths):
            if not pattern.search(x):
                chroms[x] = l
            else:
                removed.append(x)

    logging.info("Chromosomes included in analysis: {}".format(", ".join(natsorted(chroms))))
    if removed:
        logging.info("Chromosomes removed from analysis: {} (and {} more.)\n".format(", ".join(natsorted(removed[:5])), max(0, len(removed) - 5)))
    else:
        logging.info("No chromosomes matching removal pattern: {}\n".format(remove_pattern))


    return chroms

def egl_and_chromsizes(args):

    read_length = find_readlength(args)

    if args["autodetect_chroms"]:
        chromsizes = get_chroms_from_bam([f for f in args["treatment"] if f.endswith(".bam")][0], args["discard_chromosomes_pattern"])
        egf = get_effective_genome_fraction(args["genome"], read_length)
        genome_length = sum(chromsizes.values())
        egl = egf * genome_length

    else:

        have_chromsizes = args["chromsizes"] != None
        have_effective_genome_fraction = args["effective_genome_fraction"] != None

        if have_chromsizes and have_effective_genome_fraction:
            # assert have_chromsizes
            # assert have_effective_genome_fraction

            chromsizes = create_genome_size_dict_custom_genome(args["chromsizes"])
            egf = args["effective_genome_fraction"]
            egl = egf * sum(chromsizes.values())
            genome_length = sum(chromsizes.values())
        elif have_chromsizes:
            chromsizes = create_genome_size_dict_custom_genome(args["chromsizes"])
            egf = get_effective_genome_fraction(args["genome"], read_length)
            egl = egf * sum(chromsizes.values())
            genome_length = sum(chromsizes.values())
        elif have_effective_genome_fraction:
            chromsizes = create_genome_size_dict(args["genome"])
            egf = args["effective_genome_fraction"]
            genome_length = sum(chromsizes.values())
            egl = egf * genome_length

        else:
            chromsizes = create_genome_size_dict(args["genome"])
            egf = get_effective_genome_fraction(args["genome"], read_length)
            genome_length = sum(chromsizes.values())
            egl = egf * genome_length

    logging.info("Using genome {}.\n".format(args["genome"]))
    logging.info("Using an effective genome length of ~{} * 1e6\n".format(int(egl/1e6)))

    return egl, chromsizes
